from __future__ import annotations

import logging
import itertools as it
from typing import TYPE_CHECKING
from collections import OrderedDict

from .abc import (
    Crosser,
)

if TYPE_CHECKING:
    from typing import (
        Optional,
        Dict,
    )
    from ....models import (
        PlannedTrip,
        Route,
        Trip,
    )

logger = logging.getLogger(__name__)


class OrderedCrosser(Crosser):
    ranking: Dict[Route, OrderedDict[Trip, PlannedTrip]]

    def __init__(self, neighborhood_max_size: int = 250, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if neighborhood_max_size is None:
            neighborhood_max_size = len(self.trips)
        else:
            neighborhood_max_size = min(neighborhood_max_size, len(self.trips))

        self.neighborhood_max_size = neighborhood_max_size
        self.ranking = dict()
        self.initialize_ranking()

    def initialize_ranking(self) -> None:
        logger.info("Initializing ranking...")
        for route in self.routes:
            self.ranking[route] = self.create_sub_ranking(route)
            logger.info(f"Added sub ranking! Currently {len(self.ranking)}.")

    def create_sub_ranking(self, route: Route) -> OrderedDict[Trip, PlannedTrip]:
        logger.debug("Creating sub_ranking...")
        raw_sub_ranking = route.conjecture_trip_in_batch(it.islice(self.pending_trips, self.neighborhood_max_size))

        self.criterion.sorted(raw_sub_ranking, inplace=True)
        return OrderedDict((item.trip, item) for item in raw_sub_ranking)

    def update_ranking(self, planned_trip: PlannedTrip) -> None:
        logger.info("Updating ranking...")
        for route in self.ranking:
            self.ranking[route].pop(planned_trip.trip, None)
        self.ranking[planned_trip.route] = self.create_sub_ranking(planned_trip.route)

    def mark_planned_trip_as_done(self, planned_trip: PlannedTrip):
        super().mark_planned_trip_as_done(planned_trip)
        self.update_ranking(planned_trip)

    def get_planned_trip(self) -> Optional[PlannedTrip]:
        if len(self.ranking) == 0:
            return None

        best = None
        for sub_ranking in self.ranking.values():
            if len(sub_ranking) == 0:
                continue
            current = next(iter(sub_ranking.values()))
            if not current.feasible:
                continue
            best = self.criterion.best(best, current)
        return best



import logging

from .abc import (
    Breeder,
)
from ....models import (
    Result
)

logger = logging.getLogger(__name__)


class FlipBreeder(Breeder):

    def improve(self) -> Result:
        for route in self.planning:
            cost = self.objective.optimization_function(route)

            for i in range(1, len(route.stops) - 1):
                j = i + 1
                first = route.stops[i]
                second = route.stops[j]

                if not set(first.pickups).isdisjoint(second.deliveries):
                    continue
                first.flip(second)

                if not route.feasible or cost < self.objective.optimization_function(route):
                    second.flip(first)
                    continue

                cost = self.objective.optimization_function(route)
                logger.info(f'Flipped "{i}"-th and "{j}"-th stops from "{route}".')
        return self.result



from __future__ import annotations

import logging
from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    TYPE_CHECKING,
)

from ....exceptions import (
    StopPlannedTripIterationException,
)
from ....models import (
    Route,
    ShortestTimePlannedTripCriterion,
)

if TYPE_CHECKING:
    from typing import (
        Set,
        Optional,
        Type,
    )
    from ....models import (
        PlannedTrip,
        Vehicle,
        Fleet,
        Trip,
        Job,
        PlannedTripCriterion,
    )

logger = logging.getLogger(__name__)


class Crosser(ABC):
    fleet: Fleet
    job: Job
    criterion_cls: Type[PlannedTripCriterion]
    _criterion: Optional[PlannedTripCriterion]

    def __init__(self, fleet: Fleet, job: Job, criterion_cls: Type[PlannedTripCriterion] = None, *args, **kwargs):
        if criterion_cls is None:
            criterion_cls = ShortestTimePlannedTripCriterion

        self.fleet = fleet
        self.job = job
        self.routes = set(Route(vehicle) for vehicle in self.vehicles)
        self._pending_trips = set(self.trips)

        self.criterion_cls = criterion_cls
        self._criterion = None

        self.args = args
        self.kwargs = kwargs

    @property
    def criterion(self) -> PlannedTripCriterion:
        if self._criterion is None:
            self._criterion = self.criterion_cls(*self.args, **self.kwargs)
        return self._criterion

    @property
    def vehicles(self) -> Set[Vehicle]:
        return self.fleet.vehicles

    @property
    def trips(self) -> Set[Trip]:
        return self.job.trips

    def __iter__(self):
        return self

    def __next__(self):
        while True:
            planned_trip = self.get_planned_trip()
            if planned_trip is None:
                raise StopPlannedTripIterationException()
            logger.debug(f'Yielding {planned_trip}...')
            return planned_trip

    @abstractmethod
    def get_planned_trip(self) -> Optional[PlannedTrip]:
        pass

    def mark_planned_trip_as_done(self, planned_trip: PlannedTrip) -> None:
        self.mark_trip_as_done(planned_trip.trip)

    def mark_trip_as_done(self, trip: Trip) -> None:
        self._pending_trips.remove(trip)

    def mark_planned_trip_as_undone(self, planned_trip: PlannedTrip) -> None:
        self.mark_trip_as_undone(planned_trip.trip)

    def mark_trip_as_undone(self, trip: Trip) -> None:
        self._pending_trips.add(trip)

    @property
    def pending_trips(self) -> Set[Trip]:
        return self._pending_trips

    @property
    def done_trips(self) -> Set[Trip]:
        return self.trips - self.pending_trips

    @property
    def completed(self) -> bool:
        return len(self.pending_trips) == 0



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from .constants import (
    OptimizationDirection,
    MIN_FLOAT,
    MAX_FLOAT,
)

if TYPE_CHECKING:
    from typing import (
        Optional,
        List,
        Iterable,
    )
    from .planned_trips import (
        PlannedTrip,
    )

logger = logging.getLogger(__name__)


class PlannedTripCriterion(ABC):
    def __init__(self, name: str, direction: OptimizationDirection, *args, **kwargs):
        self.name = name
        self.direction = direction

    @abstractmethod
    def scoring(self, planned_trip: PlannedTrip) -> float:
        pass

    def best(self, *args: Optional[PlannedTrip]) -> PlannedTrip:
        return self.direction(
            (arg for arg in args if arg is not None),
            key=self.scoring,
            default=None,
        )

    def sorted(self, planned_trips: Iterable[PlannedTrip], inplace: bool = False) -> List[PlannedTrip]:
        return self.direction.sorted(planned_trips, key=self.scoring, inplace=inplace)

    def nbest(self, n: int, planned_trips: Iterable[PlannedTrip], inplace: bool = False):
        return self.direction.nbest(n, planned_trips, key=self.scoring, inplace=inplace)


class ShortestTimePlannedTripCriterion(PlannedTripCriterion):

    def __init__(self, *args, **kwargs):
        super().__init__(
            direction=OptimizationDirection.MINIMIZATION,
            name='Shortest-Time',
            *args, **kwargs,
        )

    def scoring(self, planned_trip: PlannedTrip) -> float:
        if not planned_trip.feasible:
            return MAX_FLOAT

        return planned_trip.delivery_time - planned_trip.route.last_departure_time


class LongestTimePlannedTripCriterion(PlannedTripCriterion):

    def __init__(self, *args, **kwargs):
        super().__init__(
            direction=OptimizationDirection.MAXIMIZATION,
            name='Longest-Time',
            *args, **kwargs,
        )

    def scoring(self, planned_trip: PlannedTrip) -> float:
        if not planned_trip.feasible:
            return MIN_FLOAT

        return planned_trip.delivery_time - planned_trip.route.last_departure_time


class LongestUtilTimePlannedTripCriterion(PlannedTripCriterion):

    def __init__(self, *args, **kwargs):
        super().__init__(
            direction=OptimizationDirection.MAXIMIZATION,
            name='Longest-Time',
            *args, **kwargs,
        )

    def scoring(self, planned_trip: PlannedTrip) -> float:
        if not planned_trip.feasible:
            return MIN_FLOAT

        return planned_trip.duration - planned_trip.trip.origin.distance_to(planned_trip.route.last_position)


class HashCodePlannedTripCriterion(PlannedTripCriterion):

    def __init__(self, *args, **kwargs):
        super().__init__(
            direction=OptimizationDirection.MAXIMIZATION,
            name='Longest-Time',
            *args, **kwargs,
        )

    def scoring(self, planned_trip: PlannedTrip) -> float:
        if not planned_trip.feasible:
            return MIN_FLOAT

        scoring = planned_trip.distance
        if planned_trip.pickup_time == planned_trip.trip.earliest:
            scoring += planned_trip.trip.on_time_bonus

        # TODO: Optimize this call
        scoring -= planned_trip.origin.distance_to(planned_trip.route.last_position)

        return scoring



from __future__ import annotations

from abc import (
    ABC,
    abstractmethod,
)
from copy import deepcopy

from ....models import (
    Planning,
    Result,
    Objective,
)


class Breeder(ABC):

    def __init__(self, result: Result, with_copy: bool = True):
        if with_copy:
            result = deepcopy(result)

        self.result = result

    @property
    def objective(self) -> Objective:
        return self.result.objective

    @property
    def planning(self) -> Planning:
        return self.result.planning

    @abstractmethod
    def improve(self) -> Result:
        pass



import unittest
from copy import deepcopy

import jinete as jit
from tests.utils import (
    generate_one_surface,
    generate_one_position,
)


class TestGeometricPosition(unittest.TestCase):

    def test_construction(self):
        surface = generate_one_surface()
        position = jit.GeometricPosition([3, 4], surface)
        self.assertEqual(3, position[0])
        self.assertEqual(4, position[1])

    def test_deepcopy(self):
        position = generate_one_position()
        copied_position = deepcopy(position)
        self.assertEqual(position, copied_position)


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from abc import ABC

from typing import (
    TYPE_CHECKING,
    TypeVar)
from .constants import (
    OptimizationDirection,
)

from .routes import (
    Route,
)
from .plannings import (
    Planning,
)
from .planned_trips import (
    PlannedTrip,
)
from .results import (
    Result,
)
from .stops import (
    Stop,
)

if TYPE_CHECKING:
    from typing import (
        Optional
    )

    Optimizable = TypeVar('Optimizable', Result, Planning, Route, Stop, PlannedTrip)


class Objective(ABC):
    direction: OptimizationDirection

    def __init__(self, name: str, direction: OptimizationDirection):
        self.name = name
        self.direction = direction

    def best(self, *args: Optional[Optimizable]) -> Optimizable:
        return self.direction(
            (arg for arg in args if arg is not None),
            key=self.optimization_function,
            default=None,
        )

    def optimization_function(self, value: Optimizable) -> float:
        if isinstance(value, Result):
            return self._result_optimization_function(value)
        elif isinstance(value, Planning):
            return self._planning_optimization_function(value)
        elif isinstance(value, Route):
            return self._route_optimization_function(value)
        elif isinstance(value, Stop):
            return self._stop_optimization_function(value)
        elif isinstance(value, PlannedTrip):
            return self._planned_trip_optimization_function(value)
        else:
            raise ValueError

    def _result_optimization_function(self, result: Result) -> float:
        return self._planning_optimization_function(result.planning)

    def _planning_optimization_function(self, planning: Planning) -> float:
        scoring = 0.0
        for route in planning:
            scoring += self._route_optimization_function(route)
        return scoring

    def _route_optimization_function(self, route: Route) -> float:
        scoring = 0.0
        for stop in route.stops:
            scoring += self._stop_optimization_function(stop)
        return scoring

    def _stop_optimization_function(self, stop: Stop) -> float:
        scoring = 0.0
        for planned_trip in stop.pickups:
            scoring += self._planned_trip_optimization_function(planned_trip)
        return scoring

    def _planned_trip_optimization_function(self, planned_trip: PlannedTrip) -> float:
        raise NotImplementedError


class DialARideObjective(Objective):

    def __init__(self):
        super().__init__(
            name='Dial-a-Ride',
            direction=OptimizationDirection.MINIMIZATION,
        )

    def _route_optimization_function(self, route: Route) -> float:
        scoring = 0.0
        for stop in route.stops:
            scoring += stop.distance
        return scoring

    def _planned_trip_optimization_function(self, planned_trip: PlannedTrip) -> float:
        scoring = 0.0
        current = planned_trip.pickup
        while current != planned_trip.delivery and current.following is not None:
            current = current.following
            scoring += current.distance
        return scoring


class TaxiSharingObjective(Objective):

    def __init__(self):
        super().__init__(
            name='Taxi-Sharing',
            direction=OptimizationDirection.MAXIMIZATION,
        )

    def _planned_trip_optimization_function(self, planned_trip: PlannedTrip) -> float:
        if planned_trip.capacity == 0:
            return 0.0
        return planned_trip.duration


class HashCodeObjective(Objective):
    def __init__(self):
        super().__init__(
            name='HashCode-2018',
            direction=OptimizationDirection.MAXIMIZATION,
        )

    def _planned_trip_optimization_function(self, planned_trip: PlannedTrip) -> float:
        if planned_trip.capacity == 0:
            return 0.0
        trip = planned_trip.trip
        scoring = trip.distance
        if trip.earliest == planned_trip.pickup_time:
            scoring += trip.on_time_bonus
        return scoring



from __future__ import annotations

import unittest
from typing import TYPE_CHECKING
import jinete as jit

if TYPE_CHECKING:
    from typing import (
        List,
    )


class TestLongestTimePlannedTripCriterion(unittest.TestCase):
    planned_trips: List[jit.PlannedTrip]

    @classmethod
    def setUpClass(cls) -> None:
        surface = jit.GeometricSurface(jit.DistanceMetric.MANHATTAN)
        vehicle = jit.Vehicle(
            identifier='TEST',
            initial=surface.get_or_create_position([0, 0]),
        )
        route = jit.Route(vehicle)
        pickup_stop_1 = jit.Stop(route, surface.get_or_create_position([0, 0]), route.last_stop)
        delivery_stop_1 = jit.Stop(route, surface.get_or_create_position([1, 1]), pickup_stop_1)

        pickup_stop_2 = jit.Stop(route, surface.get_or_create_position([0, 0]), route.last_stop)
        delivery_stop_2 = jit.Stop(route, surface.get_or_create_position([10, 10]), pickup_stop_2)

        cls.planned_trips = [
            jit.PlannedTrip(
                route=route,
                trip=jit.Trip(
                    identifier='TEST_1',
                    origin=surface.get_or_create_position([0, 0]),
                    destination=surface.get_or_create_position([1, 1]),
                    earliest=0.0,
                    timeout=10.0,
                ),
                pickup=pickup_stop_1,
                delivery=delivery_stop_1,
            ),
            jit.PlannedTrip(
                route=route,
                trip=jit.Trip(
                    identifier='TEST_1',
                    origin=surface.get_or_create_position([1, 1]),
                    destination=surface.get_or_create_position([10, 10]),
                    earliest=0.0,
                    timeout=20.0,
                ),
                pickup=pickup_stop_2,
                delivery=delivery_stop_2,
            )
        ]

    def test_creation(self):
        criterion = jit.LongestTimePlannedTripCriterion()
        self.assertEqual(jit.OptimizationDirection.MAXIMIZATION, criterion.direction)
        self.assertEqual('Longest-Time', criterion.name)

    def test_sorting(self):
        criterion = jit.LongestTimePlannedTripCriterion()

        self.assertEqual(
            self.planned_trips[::-1],
            criterion.sorted(self.planned_trips),
        )

    def test_scoring(self):
        criterion = jit.LongestTimePlannedTripCriterion()

        self.assertEqual(
            2.0,
            criterion.scoring(self.planned_trips[0]),
        )

        self.assertEqual(
            20.0,
            criterion.scoring(self.planned_trips[1]),
        )

    def test_best(self):
        criterion = jit.LongestTimePlannedTripCriterion()

        self.assertEqual(
            self.planned_trips[1],
            criterion.best(*self.planned_trips),
        )


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

from typing import TYPE_CHECKING

from .abc import (
    Model,
)
from .objectives import (
    Objective,
    DialARideObjective,
)

if TYPE_CHECKING:
    from typing import (
        Set,
        Any,
        Dict,
        Type,
        Optional,
    )
    from .trips import (
        Trip,
    )


class Job(Model):
    trips: Set[Trip]
    objective_cls: Type[Objective]
    _objective: Optional[Objective]

    def __init__(self, trips: Set[Trip], objective_cls: Type[Objective], *args, **kwargs):
        if objective_cls is None:
            objective_cls = DialARideObjective

        self.trips = trips
        self.objective_cls = objective_cls
        self._objective = None

        self.args = args
        self.kwargs = kwargs

    @property
    def objective(self) -> Objective:
        if self._objective is None:
            self._objective = self.objective_cls(*self.args, **self.kwargs)
        return self._objective

    def __iter__(self):
        yield from self.trips

    def __deepcopy__(self, memo) -> Job:
        return self

    def as_dict(self) -> Dict[str, Any]:
        trips_str = ', '.join(str(trip) for trip in self.trips)
        dict_values = {
            'trips': f'{{{trips_str}}}',
            'objective': self.objective,
        }
        return dict_values



from __future__ import annotations

import logging
from typing import (
    TYPE_CHECKING,
    Optional,
)
from .constants import (
    MAX_FLOAT,
)

if TYPE_CHECKING:
    from typing import (
        Dict,
        Any,
    )
    from .positions import (
        Position,
    )

logger = logging.getLogger(__name__)


class Trip(object):
    __slots__ = (
        'identifier',
        'origin',
        'destination',
        'earliest',
        'timeout',
        'on_time_bonus',
        'load_capacity',
        'load_time',
        'inbound',
        'capacity',
    )
    identifier: str
    origin: Position
    destination: Position
    earliest: float
    timeout: Optional[float]
    on_time_bonus: float
    load_time: float
    capacity: float

    def __init__(self, identifier: str, origin: Position, destination: Position, earliest: float = 0.0,
                 timeout: Optional[float] = None, on_time_bonus: float = 0.0, load_time: float = 0.0,
                 inbound: bool = True, capacity: float = 1):
        self.identifier = identifier
        self.origin = origin
        self.destination = destination
        self.earliest = earliest
        self.timeout = timeout
        self.on_time_bonus = on_time_bonus
        self.load_time = load_time
        self.inbound = inbound
        self.capacity = capacity

    @property
    def latest(self) -> float:
        if self.timeout is None:
            return MAX_FLOAT
        return self.earliest + self.timeout

    @property
    def empty(self) -> bool:
        return self.capacity == 0

    @property
    def distance(self) -> float:
        return self.origin.distance_to(self.destination)

    def duration(self, now: float):
        return self.origin.time_to(self.destination, now)

    def __deepcopy__(self, memo: Dict[int, Any]) -> Trip:
        return self



import unittest

import jinete as jit

from tests.utils import (
    generate_one_planned_trip,
    generate_one_vehicle,
)


class TestPlannedTrip(unittest.TestCase):

    def test_feasible_planned_trip(self):
        vehicle = generate_one_vehicle()
        route = jit.Route(vehicle)

        planned_trip = generate_one_planned_trip(True, route)
        self.assertTrue(planned_trip.feasible)

    def test_not_feasible_planned_trip(self):
        vehicle = generate_one_vehicle()
        route = jit.Route(vehicle)

        planned_trip = generate_one_planned_trip(False, route)
        self.assertFalse(planned_trip.feasible)


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

import unittest
from copy import deepcopy
from typing import TYPE_CHECKING

import jinete as jit

from tests.utils import (
    generate_one_vehicle,
    generate_one_position,
    generate_one_route,
)

if TYPE_CHECKING:
    from typing import (
        List,
    )


class TestStop(unittest.TestCase):
    vehicle: jit.Vehicle
    route: jit.Route
    position: jit.Position
    stops: List[jit.Stop]

    @classmethod
    def setUpClass(cls) -> None:
        cls.vehicle = generate_one_vehicle()
        cls.route = jit.Route(cls.vehicle)

        cls.position = generate_one_position()

    def setUp(self) -> None:
        stop0 = jit.Stop(self.route, self.position, None)

        stop1 = jit.Stop(self.route, generate_one_position(), stop0)
        stop0.following = stop1

        stop2 = jit.Stop(self.route, generate_one_position(), stop1)
        stop1.following = stop2

        stop3 = jit.Stop(self.route, generate_one_position(), stop2)
        stop2.following = stop3

        stop4 = jit.Stop(self.route, generate_one_position(), stop3)
        stop3.following = stop4

        stops = [stop0, stop1, stop2, stop3, stop4]
        self.stops = stops
        self.route.stops = stops

    def test_creation(self):
        stop = jit.Stop(self.route, self.position, None)

        self.assertIsNone(stop._down_time)
        self.assertIsNone(stop._load_time)
        self.assertIsNone(stop._earliest)
        self.assertIsNone(stop._arrival_time)

        self.assertEqual(stop.route, self.route)
        self.assertEqual(stop.position, self.position)
        self.assertEqual(stop.previous, None)
        self.assertEqual(stop.previous_position, self.vehicle.initial)
        self.assertEqual(stop.previous_departure_time, self.vehicle.earliest)
        self.assertEqual(
            stop.navigation_time,
            stop.position.time_to(self.vehicle.initial, stop.previous_departure_time),
        )
        self.assertEqual(stop.waiting_time, 0.0)
        self.assertEqual(stop.down_time, 0.0)
        self.assertEqual(stop.load_time, 0.0)

    def test_creation_with_previous(self):
        previous_position = generate_one_position()
        previous_stop = jit.Stop(self.route, previous_position, None)

        stop = jit.Stop(self.route, self.position, previous_stop)
        previous_stop.following = stop

        self.assertEqual(previous_stop.following, stop)

        self.assertEqual(stop.route, self.route)
        self.assertEqual(stop.position, self.position)
        self.assertEqual(stop.previous, previous_stop)
        self.assertEqual(stop.distance, stop.position.distance_to(previous_stop.position))
        self.assertEqual(stop.previous_position, previous_stop.position)
        self.assertEqual(stop.previous_departure_time, previous_stop.departure_time)
        self.assertEqual(
            stop.navigation_time,
            stop.position.time_to(previous_stop.position, stop.previous_departure_time),
        )

    def test_stops(self):
        route = generate_one_route()
        stop = route.stops[0]
        self.assertEqual(route.stops, stop.stops)

    def test_flush(self):
        stop = jit.Stop(self.route, self.position, None)

        self.assertIsNone(stop._down_time)
        self.assertIsNone(stop._load_time)
        self.assertIsNone(stop._earliest)
        self.assertIsNone(stop._arrival_time)

        self.assertIsInstance(stop.departure_time, float)

        self.assertIsNotNone(stop._down_time)
        self.assertIsNotNone(stop._load_time)
        self.assertIsNotNone(stop._earliest)
        self.assertIsNotNone(stop._arrival_time)

        stop.flush()

        self.assertIsNone(stop._down_time)
        self.assertIsNone(stop._load_time)
        self.assertIsNone(stop._earliest)
        self.assertIsNone(stop._arrival_time)

    def test_cache(self):
        self.assertIsInstance(self.stops[-1].departure_time, float)
        for stop in self.stops:
            self.assertIsNotNone(stop._down_time)
            self.assertIsNotNone(stop._load_time)
            self.assertIsNotNone(stop._earliest)
            self.assertIsNotNone(stop._arrival_time)

    def test_all_following(self):
        self.assertIsInstance(self.stops[-1].departure_time, float)
        self.stops[2].flush_all_following()

        for stop in self.stops[:2]:
            self.assertIsNotNone(stop._down_time)
            self.assertIsNotNone(stop._load_time)
            self.assertIsNotNone(stop._earliest)
            self.assertIsNotNone(stop._arrival_time)
        for stop in self.stops[2:]:
            self.assertIsNone(stop._down_time)
            self.assertIsNone(stop._load_time)
            self.assertIsNone(stop._earliest)
            self.assertIsNone(stop._arrival_time)

    def test_all_previous(self):
        self.assertIsInstance(self.stops[-1].departure_time, float)

        self.stops[2].flush_all_previous()

        for stop in self.stops[:3]:
            self.assertIsNone(stop._down_time)
            self.assertIsNone(stop._load_time)
            self.assertIsNone(stop._earliest)
            self.assertIsNone(stop._arrival_time)

        for stop in self.stops[3:]:
            self.assertIsNotNone(stop._down_time)
            self.assertIsNotNone(stop._load_time)
            self.assertIsNotNone(stop._earliest)
            self.assertIsNotNone(stop._arrival_time)

    def test_flip(self):
        route = deepcopy(self.route)
        stops = route.stops
        stops[1].flip(stops[2])

        self.assertEqual(stops[0].position, self.route.stops[0].position)
        self.assertEqual(stops[1].position, self.route.stops[2].position)
        self.assertEqual(stops[2].position, self.route.stops[1].position)
        self.assertEqual(stops[3].position, self.route.stops[3].position)

        for first, second in zip(stops[:-1], stops[1:]):
            self.assertEqual(first.following, second)
            self.assertEqual(second.previous, first)


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

import logging
from copy import deepcopy
from math import isnan
from typing import (
    TYPE_CHECKING,
)
from uuid import (
    uuid4,
)
from .abc import (
    Model,
)
from .planned_trips import (
    PlannedTrip,
)
from .stops import (
    Stop,
)

if TYPE_CHECKING:
    from typing import (
        List,
        Iterator,
        Any,
        Dict,
        Optional,
        Iterable,
    )
    from uuid import (
        UUID,
    )
    from .vehicles import (
        Vehicle,
    )
    from .trips import (
        Trip,
    )
    from .positions import (
        Position,
    )

logger = logging.getLogger(__name__)


class Route(Model):
    vehicle: Vehicle
    uuid: UUID
    stops: List[Stop]

    def __init__(self, vehicle: Vehicle, stops: List[Stop] = None, uuid: UUID = None):

        if uuid is None:
            uuid = uuid4()
        self.uuid = uuid

        self.vehicle = vehicle

        if stops is None:
            self.stops = [
                Stop(self, self.vehicle.initial, None),
            ]
        else:
            self.stops = list(stops)

    def __iter__(self):
        yield from self.planned_trips

    def __deepcopy__(self, memo: Dict[int, Any]) -> Route:
        vehicle = deepcopy(self.vehicle, memo)

        route = Route(vehicle)
        memo[id(self)] = route

        route.stops = deepcopy(self.stops, memo)

        return route

    @property
    def planned_trips(self) -> Iterator[PlannedTrip]:
        yield from self.deliveries

    @property
    def pickups(self) -> Iterator[PlannedTrip]:
        for stop in self.stops:
            yield from stop.pickups

    @property
    def deliveries(self) -> Iterator[PlannedTrip]:
        for stop in self.stops:
            yield from stop.deliveries

    @property
    def feasible(self) -> bool:
        if any(self.planned_trips):
            if not self.first_stop.position == self.vehicle.initial:
                return False
            if not self.vehicle.earliest <= self.first_stop.arrival_time:
                return False
            if not self.last_position == self.vehicle.final:
                return False
            if not self.last_departure_time <= self.vehicle.latest:
                return False

        if __debug__:
            for stop in self.stops:
                assert stop.route == self

            for one, two in zip(self.stops[:-1], self.stops[1:]):
                assert one.following == two
                assert two.previous == one
                assert one.position != two.position

        for planned_trip in self.planned_trips:
            if not planned_trip.feasible:
                return False
        return True

    @property
    def loaded(self):
        return any(self.planned_trips)

    @property
    def trips(self) -> Iterator[Trip]:
        yield from (planned_trip.trip for planned_trip in self.planned_trips)

    @property
    def loaded_planned_trips(self) -> Iterator[PlannedTrip]:
        yield from (planned_trip for planned_trip in self.planned_trips if not planned_trip.empty)

    @property
    def loaded_trips(self) -> Iterator[Trip]:
        yield from (planned_trip.trip for planned_trip in self.loaded_planned_trips)

    @property
    def loaded_trips_count(self) -> int:
        return sum(1 for _ in self.loaded_trips)

    @property
    def first_arrival_time(self) -> float:
        return self.first_stop.arrival_time

    @property
    def first_departure_time(self) -> float:
        return self.first_stop.departure_time

    @property
    def last_arrival_time(self) -> float:
        return self.last_stop.arrival_time

    @property
    def last_departure_time(self) -> float:
        return self.last_stop.departure_time

    @property
    def duration(self) -> float:
        return self.last_departure_time - self.first_arrival_time

    @property
    def last_position(self) -> Position:
        return self.last_stop.position

    @property
    def first_stop(self) -> Stop:
        stop = self.stops[0]
        assert stop.previous is None
        return stop

    @property
    def last_stop(self) -> Stop:
        stop = self.stops[-1]
        assert stop.following is None
        return stop

    @property
    def vehicle_uuid(self) -> Optional[UUID]:
        if self.vehicle is None:
            return None
        return self.vehicle.uuid

    def as_dict(self) -> Dict[str, Any]:
        return {
            'uuid': self.uuid,
            'vehicle_uuid': self.vehicle_uuid,
        }

    def conjecture_trip(self, trip: Trip) -> PlannedTrip:
        pickup = Stop(self, trip.origin, self.last_stop)
        delivery = Stop(self, trip.destination, pickup)
        planned_trip = PlannedTrip(route=self, trip=trip, pickup=pickup, delivery=delivery)
        return planned_trip

    def conjecture_trip_in_batch(self, iterable: Iterable[Trip]) -> List[PlannedTrip]:
        return [self.conjecture_trip(trip) for trip in iterable]

    def finish(self):
        if self.loaded and self.last_stop.position != self.vehicle.final:
            finish_stop = Stop(self, self.vehicle.final, self.last_stop)
            if not self.last_stop.position == finish_stop.position:
                self.append_stop(finish_stop)

    def append_stop(self, stop: Stop) -> None:
        assert stop.previous == self.last_stop
        assert set(stop.pickups).isdisjoint(stop.deliveries)
        if __debug__:
            for planned_trip in stop.pickups:
                assert planned_trip.pickup == stop
            for planned_trip in stop.deliveries:
                assert planned_trip.delivery == stop
        self.last_stop.following = stop
        self.stops.append(stop)

    def append_planned_trip(self, planned_trip: PlannedTrip):
        assert planned_trip.delivery is not None
        assert planned_trip.pickup is not None
        assert planned_trip.delivery.previous is not None
        assert planned_trip.pickup_time <= planned_trip.delivery_time
        assert isnan(planned_trip.duration) or planned_trip.duration > 0

        if not planned_trip.pickup == self.last_stop:
            if self.last_stop.position == planned_trip.pickup.position:
                self.last_stop.merge(planned_trip.pickup)
                planned_trip.delivery.previous = self.last_stop  # FIXME: should be inside merge?
            else:
                self.append_stop(planned_trip.pickup)

        self.append_stop(planned_trip.delivery)

        logger.info(f'Append trip "{planned_trip.trip_identifier}" identifier to route "{self.uuid}".')



import unittest
from uuid import UUID

import jinete as jit

from tests.utils import (
    generate_one_position,
)


class TestVehicles(unittest.TestCase):

    def test_vehicle(self):
        initial = generate_one_position()
        identifier = str(0)
        vehicle = jit.Vehicle(identifier, initial)

        self.assertEqual(identifier, vehicle.identifier)
        self.assertEqual(1, vehicle.capacity)
        self.assertEqual(vehicle.initial, initial)
        self.assertEqual(vehicle.final, vehicle.initial)
        self.assertEqual(vehicle.earliest, 0)
        self.assertEqual(vehicle.latest, jit.MAX_FLOAT)
        self.assertEqual(vehicle.timeout, jit.MAX_FLOAT)
        self.assertIsInstance(vehicle.uuid, UUID)

    def test_vehicle_with_capacity(self):
        capacity = 3
        initial = generate_one_position()
        identifier = str(0)
        vehicle = jit.Vehicle(identifier, initial, capacity=capacity)

        self.assertEqual(vehicle.capacity, capacity)
        self.assertEqual(vehicle.initial, initial)
        self.assertEqual(vehicle.final, vehicle.initial)
        self.assertEqual(vehicle.earliest, 0)
        self.assertEqual(vehicle.latest, jit.MAX_FLOAT)
        self.assertEqual(vehicle.timeout, jit.MAX_FLOAT)
        self.assertIsInstance(vehicle.uuid, UUID)

    def test_vehicle_with_final(self):
        capacity = 3
        initial = generate_one_position()
        final = generate_one_position()
        identifier = str(0)
        vehicle = jit.Vehicle(identifier, initial, capacity=capacity, final=final)

        self.assertEqual(identifier, vehicle.identifier)
        self.assertEqual(vehicle.capacity, capacity)
        self.assertEqual(vehicle.initial, initial)
        self.assertEqual(vehicle.final, final)
        self.assertEqual(vehicle.earliest, 0)
        self.assertEqual(vehicle.latest, jit.MAX_FLOAT)
        self.assertEqual(vehicle.timeout, jit.MAX_FLOAT)
        self.assertIsInstance(vehicle.uuid, UUID)

    def test_vehicle_with_earliest(self):
        initial = generate_one_position()
        earliest = 3600
        identifier = str(0)
        vehicle = jit.Vehicle(identifier, initial, earliest=earliest)

        self.assertEqual(identifier, vehicle.identifier)
        self.assertEqual(vehicle.capacity, 1, )
        self.assertEqual(vehicle.initial, initial)
        self.assertEqual(vehicle.final, vehicle.initial)
        self.assertEqual(vehicle.earliest, earliest)
        self.assertEqual(vehicle.latest, jit.MAX_FLOAT)
        self.assertEqual(vehicle.timeout, jit.MAX_FLOAT)
        self.assertIsInstance(vehicle.uuid, UUID)

    def test_vehicle_with_timeout(self):
        initial = generate_one_position()
        earliest = 1800
        timeout = 3600
        identifier = str(0)
        vehicle = jit.Vehicle(identifier, initial, earliest=earliest, timeout=timeout)

        self.assertEqual(identifier, vehicle.identifier)
        self.assertEqual(vehicle.capacity, 1)
        self.assertEqual(vehicle.initial, initial)
        self.assertEqual(vehicle.final, vehicle.initial)
        self.assertEqual(vehicle.earliest, earliest)
        self.assertEqual(vehicle.latest, earliest + timeout)
        self.assertEqual(vehicle.timeout, timeout)
        self.assertIsInstance(vehicle.uuid, UUID)

    def test_as_dict(self):
        initial = generate_one_position()
        identifier = str(0)
        vehicle = jit.Vehicle(identifier, initial)

        expected = {
            'initial': vehicle.initial,
            'final': vehicle.final,
            'capacity': vehicle.capacity,
            'earliest': vehicle.earliest,
            'timeout': vehicle.timeout,
            'route_timeout': vehicle.route_timeout,
            'trip_timeout': vehicle.trip_timeout,
            'latest': vehicle.latest,
            'uuid': vehicle.uuid,
        }

        self.assertEqual(expected, vehicle.as_dict())

if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from ...models import (
    GeometricSurface,
    DistanceMetric,
    Fleet,
    Vehicle,
    Job,
    Trip,
    DialARideObjective,
)
from .exceptions import (
    LoaderFormatterException,
)
from .abc import (
    LoaderFormatter,
)

if TYPE_CHECKING:
    from ...models import (
        Surface,
    )

logger = logging.getLogger(__name__)


class CordeauLaporteLoaderFormatter(LoaderFormatter):

    def fleet(self, surface: Surface, *args, **kwargs) -> Fleet:
        row = self.data[0]
        m = int(row[0])

        depot_row = self.data[1]
        initial = surface.get_or_create_position(depot_row[1:3])
        final = None

        capacity = row[3]
        route_timeout = row[2]
        trip_timeout = row[4]

        vehicles = set()
        for idx in range(m):
            vehicle = Vehicle(
                str(idx),
                initial,
                final,
                capacity=capacity,
                route_timeout=route_timeout,
                trip_timeout=trip_timeout,
            )

            vehicles.add(vehicle)
        fleet = Fleet(vehicles)
        logger.info(f'Created fleet!')
        return fleet

    def job(self, surface: Surface, *args, **kwargs) -> Job:
        row = self.data[0]
        n = int(row[1] // 2)

        trips = set()
        for idx in range(n):
            trip = self.build_trip(surface, idx, n)
            trips.add(trip)
        job = Job(trips, objective_cls=DialARideObjective)
        logger.info(f'Created job!')
        return job

    def build_trip(self, surface: Surface, idx: int, n: int) -> Trip:
        origin_idx = idx + 2
        destination_idx = origin_idx + n

        origin_row = self.data[origin_idx]
        destination_row = self.data[destination_idx]

        origin = surface.get_or_create_position(origin_row[1:3])
        destination = surface.get_or_create_position(destination_row[1:3])

        e1, l1 = origin_row[5:7]
        e2, l2 = destination_row[5:7]

        if e1 == 0 and l1 == 1440:
            earliest, latest = e2, l2
            inbound = True
        elif e2 == 0 and l2 == 1440:
            earliest, latest = e1, l1
            inbound = False
        else:
            raise LoaderFormatterException('It is not possible to distinguish between inbound and outbound task.')

        identifier = str(idx)
        timeout = latest - earliest

        trip = Trip(
            identifier=identifier,
            origin=origin,
            destination=destination,
            inbound=inbound,
            earliest=earliest,
            timeout=timeout,
            load_time=10.0,
        )
        return trip

    def surface(self, *args, **kwargs) -> Surface:
        surface = GeometricSurface(DistanceMetric.EUCLIDEAN)
        logger.info(f'Created surface!')
        return surface



from __future__ import annotations

import logging
from copy import deepcopy
from typing import TYPE_CHECKING
from uuid import (
    uuid4,
)
from .abc import (
    Model,
)

if TYPE_CHECKING:
    from typing import (
        Set,
        Dict,
        Any,
        Iterator,
    )
    from .routes import (
        Route,
    )
    from .trips import (
        Trip,
    )
    from .planned_trips import (
        PlannedTrip,
    )
    from uuid import (
        UUID,
    )

logger = logging.getLogger(__name__)


class Planning(Model):
    routes: Set[Route]
    uuid: UUID

    def __init__(self, routes: Set[Route] = None, uuid: UUID = None):
        if uuid is None:
            uuid = uuid4()
        if routes is None:
            routes = set()

        self.routes = routes
        self.uuid = uuid

    def __iter__(self):
        yield from self.routes

    @property
    def loaded_routes(self):
        return set(route for route in self.routes if route.loaded)

    @property
    def planned_trips(self) -> Iterator[PlannedTrip]:
        for route in self.routes:
            yield from route.planned_trips

    @property
    def trips(self) -> Iterator[Trip]:
        yield from (planned_trip.trip for planned_trip in self.planned_trips)

    def as_dict(self) -> Dict[str, Any]:
        return {
            'uuid': self.uuid,
        }

    def __deepcopy__(self, memo: Dict[int, Any]) -> Planning:
        planning = Planning()
        memo[id(self)] = planning

        planning.routes = deepcopy(self.routes, memo)

        return planning



from __future__ import annotations

import logging
from typing import TYPE_CHECKING
import itertools as it
from .abc import (
    Crosser,
)

if TYPE_CHECKING:
    from typing import (
        Set,
        Optional,
        Iterator,
    )
    from ....models import (
        PlannedTrip,
        Trip,
        Route,
    )

logger = logging.getLogger(__name__)


class StatelessCrosser(Crosser):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._iterator = None

    def flush(self):
        self._iterator = None

    @property
    def iterator(self):
        if self._iterator is None:
            self._iterator = self._calculate_iterator()
        return self._iterator

    def _calculate_iterator(self) -> Iterator[PlannedTrip]:
        for route, trip in it.product(self.attractive_routes, self.pending_trips):
            logger.debug(f'Yielding ({route}, {trip})...')
            planned_trip = route.conjecture_trip(trip)
            yield planned_trip

    def get_planned_trip(self) -> Optional[PlannedTrip]:
        return next(self.iterator, None)

    @property
    def attractive_routes(self) -> Set[Route]:
        routes = set(route for route in self.routes if any(route.planned_trips))
        empty_route = next((route for route in self.routes if not any(route.planned_trips)), None)
        if empty_route is not None:
            routes.add(empty_route)
        return routes

    def mark_trip_as_done(self, trip: Trip):
        super().mark_trip_as_done(trip)
        self.flush()



import unittest

import jinete as jit


class TestDialARideObjective(unittest.TestCase):
    planned_trip: jit.PlannedTrip
    stop: jit.Stop
    route: jit.Route
    planning: jit.Planning
    result: jit.Result

    @classmethod
    def setUpClass(cls) -> None:
        surface = jit.GeometricSurface(jit.DistanceMetric.MANHATTAN)
        vehicle = jit.Vehicle(
            identifier='TEST',
            initial=surface.get_or_create_position([0, 0]),
        )
        fleet = jit.Fleet({vehicle})
        route = jit.Route(vehicle)

        trips = [
            jit.Trip(
                identifier='TEST_1',
                origin=surface.get_or_create_position([0, 0]),
                destination=surface.get_or_create_position([1, 1]),
                earliest=0.0,
                timeout=10.0,
            ), jit.Trip(
                identifier='TEST_1',
                origin=surface.get_or_create_position([1, 1]),
                destination=surface.get_or_create_position([10, 10]),
                earliest=0.0,
                timeout=20.0,
            ),
        ]
        job = jit.Job(set(trips), jit.DialARideObjective)

        cls.planned_trip = route.conjecture_trip(trips[0])
        cls.stop = route.stops[0]
        route.append_planned_trip(cls.planned_trip)
        route.append_planned_trip(route.conjecture_trip(trips[1]))

        cls.route = route
        cls.planning = jit.Planning({route})
        cls.result = jit.Result(fleet, job, jit.NaiveAlgorithm, cls.planning, 0.0)

    def test_creation(self):
        objective = jit.DialARideObjective()
        self.assertEqual(jit.OptimizationDirection.MINIMIZATION, objective.direction)
        self.assertEqual('Dial-a-Ride', objective.name)

    def test_result(self):
        objective = jit.DialARideObjective()

        self.assertEqual(
            20.0,
            objective.optimization_function(self.result),
        )

    def test_planning(self):
        objective = jit.DialARideObjective()

        self.assertEqual(
            20.0,
            objective.optimization_function(self.planning),
        )

    def test_route(self):
        objective = jit.DialARideObjective()

        self.assertEqual(
            20.0,
            objective.optimization_function(self.route),
        )

    def test_stop(self):
        objective = jit.DialARideObjective()

        self.assertEqual(
            2.0,
            objective.optimization_function(self.stop),
        )

    def test_planning_trip(self):
        objective = jit.DialARideObjective()

        self.assertEqual(
            2.0,
            objective.optimization_function(self.planned_trip),
        )

    def test_best(self):
        objective = jit.DialARideObjective()

        self.assertEqual(
            next(self.route.planned_trips),
            objective.best(*self.route.planned_trips),
        )


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

import logging
from typing import TYPE_CHECKING
from .abc import (
    Model,
)
from .trips import (
    Trip,
)
from .stops import (
    Stop,
)

if TYPE_CHECKING:
    from typing import (
        Dict,
        Any,
        Optional,
    )
    from uuid import (
        UUID,
    )
    from .routes import (
        Route,
    )
    from .positions import (
        Position,
    )
    from .vehicles import (
        Vehicle,
    )

logger = logging.getLogger(__name__)


class PlannedTrip(Model):
    __slots__ = [
        'route',
        'trip',
        'down_time',
        'pickup',
        'delivery',
        '_feasible',
    ]

    route: Route
    trip: Trip
    down_time: float

    _feasible: Optional[bool]

    def __init__(self, route: Route, trip: Trip, pickup: Stop, delivery: Stop, down_time: float = 0.0):
        self.route = route
        self.trip = trip
        self.down_time = down_time

        self.pickup = pickup
        pickup.append_pickup(self)

        self.delivery = delivery
        delivery.append_delivery(self)

        self._feasible = None

    @property
    def pickup_time(self) -> float:
        return self.pickup.arrival_time

    @property
    def delivery_time(self) -> float:
        return self.delivery.arrival_time

    @property
    def vehicle(self) -> Vehicle:
        return self.route.vehicle

    @property
    def trip_identifier(self) -> str:
        return self.trip.identifier

    @property
    def route_uuid(self) -> UUID:
        return self.route.uuid

    @property
    def origin(self) -> Position:
        return self.trip.origin

    @property
    def destination(self) -> Position:
        return self.trip.destination

    @property
    def distance(self) -> float:
        return self.trip.distance

    @property
    def duration(self) -> float:
        return self.delivery_time - self.pickup_time

    @property
    def capacity(self):
        return self.trip.capacity

    @property
    def feasible(self) -> bool:
        if self._feasible is None:
            self._feasible = self._calculate_feasible()
        return self._feasible

    @property
    def empty(self) -> bool:
        return self.trip.empty

    def as_dict(self) -> Dict[str, Any]:
        return {
            'route_uuid': self.route_uuid,
            'trip_identifier': self.trip_identifier,
            'pickup': self.pickup,
            'delivery': self.delivery,
            'down_time': self.down_time,
            'feasible': self.feasible,
        }

    def flush(self) -> None:
        self._feasible = None

    def _calculate_feasible(self) -> bool:
        if self.trip.inbound:
            if not self.trip.earliest <= self.delivery_time <= self.trip.latest:
                return False
        else:
            if not self.trip.earliest <= self.pickup_time <= self.trip.latest:
                return False

        time_to_return = self.trip.destination.time_to(self.vehicle.final, self.delivery_time)
        vehicle_finish_time = self.delivery_time + time_to_return
        if not vehicle_finish_time <= self.vehicle.latest:
            return False

        if self.vehicle.route_timeout is not None:
            if not self.duration <= self.vehicle.route_timeout:
                return False

        if self.vehicle.trip_timeout is not None:
            if not self.duration <= self.vehicle.trip_timeout:
                return False

        return True



from __future__ import annotations

import logging
from typing import (
    TYPE_CHECKING,
)
from .abc import (
    Model,
)

if TYPE_CHECKING:
    from uuid import (
        UUID,
    )
    from typing import (
        Dict,
        Any,
        Optional,
        Iterable,
        Tuple,
        List,
    )
    from .positions import (
        Position,
    )
    from .planned_trips import (
        PlannedTrip
    )
    from .routes import (
        Route,
    )
    from .vehicles import (
        Vehicle,
    )

logger = logging.getLogger(__name__)


class Stop(Model):
    __slots__ = [
        'route',
        'position',
        'pickups',
        'deliveries',
        'previous',
        'following',
        '_previous_departure_time',
        '_down_time',
        '_load_time',
        '_earliest',
        '_arrival_time',
    ]
    route: Route
    position: Position
    previous: Optional[Stop]
    following: Optional[Stop]
    pickups: Tuple[PlannedTrip, ...]
    deliveries: Tuple[PlannedTrip, ...]

    _down_time: Optional[float]
    _load_time: Optional[float]
    _earliest: Optional[float]
    _arrival_time: Optional[float]
    _previous_departure_time: Optional[float]

    def __init__(self, route: Route, position: Position, previous: Optional[Stop], following: Optional[Stop] = None,
                 pickups: Tuple[PlannedTrip, ...] = tuple(), deliveries: Tuple[PlannedTrip, ...] = tuple()):

        self.route = route
        self.position = position

        self.pickups = pickups
        self.deliveries = deliveries

        self.previous = previous
        self.following = following

        self._previous_departure_time = None
        self._down_time = None
        self._load_time = None
        self._earliest = None
        self._arrival_time = None

    def append_pickup(self, planned_trip: PlannedTrip) -> None:
        self.extend_pickups((planned_trip,))

    def append_delivery(self, planned_trip: PlannedTrip) -> None:
        self.extend_deliveries((planned_trip,))

    def extend_pickups(self, iterable: Iterable[PlannedTrip]) -> None:
        self.pickups = (*self.pickups, *iterable)

    def extend_deliveries(self, iterable: Iterable[PlannedTrip]) -> None:
        self.deliveries = (*self.deliveries, *iterable)

    @property
    def down_time(self) -> float:
        if self._down_time is None:
            if not any(self.pickups):
                self._down_time = 0.0
            else:
                self._down_time = max((pt.down_time for pt in self.pickups))
        return self._down_time

    @property
    def earliest(self):
        if self._earliest is None:
            if not any(self.pickups):
                self._earliest = 0.0
            else:
                self._earliest = max((pt.trip.earliest for pt in self.pickups))
        return self._earliest

    @property
    def load_time(self) -> float:
        if self._load_time is None:
            if not any(self.pickups):
                self._load_time = 0.0
            else:
                self._load_time = max((pt.trip.load_time for pt in self.pickups))
        return self._load_time

    @property
    def vehicle(self) -> Vehicle:
        return self.route.vehicle

    @property
    def vehicle_uuid(self) -> UUID:
        return self.vehicle.uuid

    @property
    def stops(self) -> List[Stop]:
        return self.route.stops

    @property
    def index(self) -> int:
        return self.stops.index(self)

    @property
    def previous_departure_time(self) -> float:
        if self.previous is None:
            return self.vehicle.earliest
        if self._previous_departure_time is None:
            self._previous_departure_time = self.previous.departure_time
        return self._previous_departure_time

    @property
    def previous_position(self) -> Position:
        if self.previous is None:
            return self.vehicle.initial
        return self.previous.position

    @property
    def distance(self) -> float:
        return self.position.distance_to(self.previous_position)

    @property
    def navigation_time(self):
        return self.previous_position.time_to(self.position, self.previous_departure_time)

    @property
    def waiting_time(self):
        return max(self.earliest - self.arrival_time, 0.0)

    @property
    def arrival_time(self):
        if self._arrival_time is None:
            arrival_time = self.previous_departure_time + self.down_time + self.navigation_time
            self._arrival_time = max(arrival_time, self.earliest)
        return self._arrival_time

    @property
    def departure_time(self) -> float:
        return self.arrival_time + self.load_time

    def as_dict(self) -> Dict[str, Any]:
        return {
            'vehicle_uuid': self.vehicle_uuid,
            'position': self.position,
        }

    def flush(self) -> None:
        self._down_time = None
        self._load_time = None
        self._earliest = None
        self._arrival_time = None

    def flush_all_previous(self):
        self.flush()
        if self.previous is not None:
            self.previous.flush_all_previous()

    def flush_all_following(self):
        self.flush()
        if self.following is not None:
            self.following.flush_all_following()

    def merge(self, other: Stop) -> None:
        if self == other:
            return
        assert self.route == other.route
        assert self.position == other.position
        assert self.following == other.following

        self.extend_pickups(other.pickups)
        for planned_trip in other.pickups:
            planned_trip.pickup = self

        self.extend_deliveries(other.deliveries)
        for planned_trip in other.deliveries:
            planned_trip.delivery = self

        self.flush()

    def flip(self, other: Stop) -> None:
        # assert other.previous == self
        assert self.route == other.route

        self_index = self.index
        other_index = other.index
        self.stops[self_index], self.stops[other_index] = self.stops[other_index], self.stops[self_index]

        following = other.following
        other.following = self
        if following is not None:
            following.previous = self
        self.following = following

        previous = self.previous
        other.previous = previous
        if previous is not None:
            previous.following = other
        self.previous = other

        other.flush_all_following()

    def flip_with_following(self):
        self.flip(self.following)



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import (
        Set,
    )
    from ..models import (
        Trip,
        Route,
        Objective,
        OptimizationDirection,
    )


class Result(object):

    def __init__(self, fleet, job, algorithm_cls, planning, computation_time):
        self.fleet = fleet
        self.job = job
        self.algorithm_cls = algorithm_cls
        self.planning = planning
        self.computation_time = computation_time

    @property
    def trips(self) -> Set[Trip]:
        return self.job.trips

    @property
    def routes(self) -> Set[Route]:
        return self.planning.routes

    @property
    def completed_trips(self) -> Set[Trip]:
        trips: Set[Trip] = set()
        for route in self.routes:
            trips |= set(route.loaded_trips)
        return trips

    @property
    def coverage_rate(self):
        return len(self.completed_trips) / len(self.job.trips)

    @property
    def objective(self) -> Objective:
        return self.job.objective

    @property
    def optimization_function(self) -> float:
        return self.objective.optimization_function(self)

    @property
    def direction(self) -> OptimizationDirection:
        return self.objective.direction
